#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "acpi.h"
#include "msg.h"

static char *conffile = NULL;

int get_config(int argc, char *argv[]) {
	/* initialise options struct */
	config.procstat = NULL;
	config.sys = NULL;
	config.acpi = NULL;
	config.daemonize = 1;
	config.loglevel = -1;
	config.use_cpufreq = 0;
	config.use_acpi = 0;
	config.max_cpu_speed = 0;
	config.min_cpu_speed = 0;
	config.cpu_states = 0;
	config.acpi_states = 0;
	config.max_speed = 0;
	config.min_speed = 0;
	config.up_threshold = 0;
	config.down_threshold = 100;
	config.ac_max_speed = 0;
	config.ac_min_speed = 0;
	config.ac_up_threshold = 0;
	config.ac_down_threshold = 100;
	config.bat_max_speed = 0;
	config.bat_min_speed = 0;
	config.bat_up_threshold = 0;
	config.bat_down_threshold = 100;

	do_test = 0;

	parse_cmdline(argc, argv);
	
	if(conffile == NULL) {
		if(parse_config_file("/etc/autothrottle.conf") == ERRVAL) return ERRVAL;
	}
	else {
		if(parse_config_file(conffile) == ERRVAL) return ERRVAL;
	}

	if(verify_config() == ERRVAL) return ERRVAL;

	check_ac();

#ifdef DEBUG
	print_config();
#endif

	return OKVAL;
}

int parse_config_file(char *name) {
	FILE *conf = NULL;
	char *tmp = (char *) calloc(256, sizeof(char)), 
		*keyword = (char *) calloc(128, sizeof(char)),
		*value = (char *) calloc(128, sizeof(char));
	int i = 0, j = 0;

	if((tmp == NULL) || (keyword == NULL) || (value == NULL)) {
		msg("Could not allocate memory.", LOG_ERROR);
		return ERRVAL;
	}

	if((conf = fopen(name, "r")) == NULL) {
		msg("Unable to open config file.", LOG_ERROR);
		return ERRVAL;
	}

	tmp = fgets(tmp, 256, conf);
	while(tmp != NULL) {
		/* read line is no comment or blank line */
		if((tmp[0] != '#') && (tmp[0] != '\n')) {
			/* Get keyword */
			i = 0;
			while((tmp[i] != '\0') && (tmp[i] != ' ') && (tmp[i] != '=') && (i < 127)) {
				keyword[i] = tmp[i];
				i++;
			}
			keyword[i] = '\0';
			if(tmp[i] == '\0') break;
			/* Skip = and blanks */
			while(((tmp[i] != '\0') && (tmp[i] == ' ')) || (tmp[i] == '=')) i++;
			if(tmp[i] == '\0') break;
			/* Get value */
			j = 0;
			while((tmp[i] != '\n') && (tmp[i] != '\0') && (i < 127)) value[j++] = tmp[i++];
			value[j] = '\0';
			
			set_value(keyword, value);
		}
		/* Get next line */
		tmp = fgets(tmp, 256, conf);
	}

	fclose(conf);
	free(tmp);
	free(keyword);
	free(value);

	return OKVAL;
}

void set_value(char *keyword, char *value) {
	char *tmp = NULL;
	/* No error checking is done here, if we're missing values, verify_config will tell
	 * us so. */
	if(!strcmp(keyword, "proc")) {
		config.procstat = (char *) calloc(strlen(value) + 6, sizeof(char));
		if(config.procstat == NULL) return;

		config.procstat = strcpy(config.procstat, value);
		tmp = config.procstat + (strlen(value) * sizeof(char));
		strcpy(tmp, "/stat");

		config.acpi = (acpifs *) calloc(1, sizeof(acpifs));
		if(config.acpi == NULL) return;

		config.acpi->throttling = (char *) calloc(strlen(value) + 31, sizeof(char));
		if(config.acpi->throttling == NULL) return;
		config.acpi->ac_state = (char *) calloc(strlen(value) + 26, sizeof(char));
		if(config.acpi->ac_state == NULL) return;

		config.acpi->throttling = strcpy(config.acpi->throttling, value);
		tmp = config.acpi->throttling + (strlen(value) * sizeof(char));
		strcpy(tmp, "/acpi/processor/CPU/throttling");

		config.acpi->ac_state = strcpy(config.acpi->ac_state, value);
		tmp = config.acpi->ac_state + (strlen(value) * sizeof(char));
		strcpy(tmp, "/acpi/ac_adapter/AC/state");
		return;
	} else if(!strcmp(keyword, "sys")) {
		config.sys = (sysfs *) calloc(1, sizeof(sysfs));
		if(config.sys == NULL) return;

		config.sys->scaling_governor = (char *) calloc(strlen(value) + 50, sizeof(char));
		if(config.sys->scaling_governor == NULL) return;
		config.sys->scaling_setspeed = (char *) calloc(strlen(value) + 50, sizeof(char));
		if(config.sys->scaling_governor == NULL) return;

		config.sys->scaling_governor = strcpy(config.sys->scaling_governor, value);
		tmp = config.sys->scaling_governor + (strlen(value) * sizeof(char));
		strcpy(tmp, "/devices/system/cpu/cpu0/cpufreq/scaling_governor");

		config.sys->scaling_setspeed = strcpy(config.sys->scaling_setspeed, value);
		tmp = config.sys->scaling_setspeed + (strlen(value) * sizeof(char));
		strcpy(tmp, "/devices/system/cpu/cpu0/cpufreq/scaling_setspeed");
		return;
	} else if(!strcmp(keyword, "loglevel")) {
		if(!strcmp(value, "info")) config.loglevel = LOG_NOTE;
		else if(!strcmp(value, "warn")) config.loglevel = LOG_WARN;
		else if(!strcmp(value, "error")) config.loglevel = LOG_ERROR;
	} else if(!strcmp(keyword, "cpufreq")) {
		if(!strcmp(value, "true")) config.use_cpufreq = 1;
		else config.use_cpufreq = 0;
		return;
	} else if(!strcmp(keyword, "acpi")) {
		if(!strcmp(value, "true")) config.use_acpi = 1;
		else config.use_acpi = 0;
		return;
	} else if(!strcmp(keyword, "max_cpu_speed")) {
		config.max_cpu_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "min_cpu_speed")) {
		config.min_cpu_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "cpu_states")) {
		config.cpu_states = atoi(value);
		return;
	} else if(!strcmp(keyword, "acpi_states")) {
		config.acpi_states = atoi(value);
		return;
	} else if(!strcmp(keyword, "ac_max_speed")) {
		config.ac_max_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "ac_min_speed")) {
		config.ac_min_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "ac_up_threshold")) {
		config.ac_up_threshold = atoi(value);
		return;
	} else if(!strcmp(keyword, "ac_down_threshold")) {
		config.ac_down_threshold = atoi(value);
		return;
	}
	else if(!strcmp(keyword, "bat_max_speed")) {
		config.bat_max_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "bat_min_speed")) {
		config.bat_min_speed = atoi(value);
		return;
	} else if(!strcmp(keyword, "bat_up_threshold")) {
		config.bat_up_threshold = atoi(value);
		return;
	} else if(!strcmp(keyword, "bat_down_threshold")) {
		config.bat_down_threshold = atoi(value);
		return;
	}
}

int verify_config(void) {
	/* To get the error messages, first check if loglevel is set to a legal
	 * value. */
	if(config.loglevel == -1) {
		config.loglevel = LOG_ERROR;
	}
	/* Check if there is anything at all to throttle. */
	if((config.use_cpufreq == 0) && (config.use_acpi == 0)) {
		msg("Neither CPUfreq nor ACPI was specified for throttling. Nothing left for me to do.", LOG_ERROR);
		return ERRVAL;
	}
	/* Check for missing variables. */
	if(config.procstat == NULL) {
		if(config.loglevel >= LOG_WARN) msg("proc not set, using default value.", LOG_WARN);
		config.procstat = (char *) calloc(11, sizeof(char));
		if(config.procstat == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}
		config.procstat = strcpy(config.procstat, "/proc/stat");

		config.acpi = (acpifs *) calloc(1, sizeof(acpifs));
		if(config.acpi == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}

		config.acpi->throttling = (char *) calloc(36, sizeof(char));
		if(config.acpi->throttling == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}
		config.acpi->ac_state = (char *) calloc(31, sizeof(char));
		if(config.acpi->ac_state == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}

		config.acpi->throttling = strcpy(config.acpi->throttling, "/proc/acpi/processor/CPU/throttling");
		config.acpi->ac_state = strcpy(config.acpi->ac_state, "/proc/acpi/ac_adapter/AC/state");
	}
	if(config.sys == NULL) {
		if(config.loglevel >= LOG_WARN) msg("sys not set, using default value.", LOG_WARN);
		config.sys = (sysfs *) calloc(1, sizeof(sysfs));
		if(config.sys == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}

		config.sys->scaling_governor = (char *) calloc(54, sizeof(char));
		if(config.sys->scaling_governor == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}
		config.sys->scaling_setspeed = (char *) calloc(54, sizeof(char));
		if(config.sys->scaling_governor == NULL) {
			msg("Could not allocate memory.", LOG_ERROR);
			return ERRVAL;
		}

		config.sys->scaling_governor = strcpy(config.sys->scaling_governor, "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor");
		config.sys->scaling_setspeed = strcpy(config.sys->scaling_setspeed, "/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed");
	}

	/* Check for legal values. */
	if((config.max_cpu_speed == 0) || (config.min_cpu_speed == 0) || (config.max_cpu_speed <= config.min_cpu_speed)) {
		msg("Please set the maximum and the minimum speed of your CPU to legal values.", LOG_ERROR);
		return ERRVAL;
	}

	if((config.use_cpufreq == 1) && (config.cpu_states == 0)) {
		msg("Please specify the amount of CPU throttling states.", LOG_ERROR);
		return ERRVAL;
	}
	if(config.use_acpi == 1) {
		if(config.acpi_states == 0) {
			msg("Please specify the amount of ACPI throttling states.", LOG_ERROR);
			return ERRVAL;
		}
		else if(config.min_cpu_speed == 0) {
			msg("Please specify the minimum speed of your CPU.", LOG_ERROR);
			return ERRVAL;
		}
	}

	if((config.ac_max_speed == 0) || (config.ac_max_speed > config.max_cpu_speed) || (config.ac_max_speed < config.ac_min_speed) ||
		(config.bat_max_speed == 0) || (config.bat_max_speed > config.max_cpu_speed) || (config.bat_max_speed < config.bat_min_speed)) {
		msg("Illegal maximum speed.", LOG_ERROR);
		return ERRVAL;
	}
	if((config.ac_min_speed <= 0) || (config.bat_min_speed <= 0)) {
		msg("Illegal minimum speed.", LOG_ERROR);
		return ERRVAL;
	}

	if((config.ac_up_threshold > 100) || (config.ac_up_threshold < config.ac_down_threshold) ||
		(config.bat_up_threshold > 100) || (config.bat_up_threshold < config.bat_down_threshold)) {
		msg("Illegal up threshold.", LOG_ERROR);
		return ERRVAL;
	}
	if((config.ac_down_threshold < 0) || (config.bat_down_threshold < 0)) {
		msg("Illegal down threshold.", LOG_ERROR);
		return ERRVAL;
	}

	return OKVAL;
}

void check_ac(void) {
	int ac = get_ac();

	if(ac == ERRVAL) {
		if(config.loglevel >= LOG_WARN) msg("Unable to determine power state. Assuming battery.", LOG_WARN);
		config.max_speed = config.bat_max_speed;
		config.min_speed = config.bat_min_speed;
		config.up_threshold = config.bat_up_threshold;
		config.down_threshold = config.bat_down_threshold;
	} else if(ac == 0) {
		if(config.loglevel == LOG_NOTE) msg("Running on battery.", LOG_NOTE);
		config.max_speed = config.bat_max_speed;
		config.min_speed = config.bat_min_speed;
		config.up_threshold = config.bat_up_threshold;
		config.down_threshold = config.bat_down_threshold;
	} else if(ac == 1) {
		if(config.loglevel == LOG_NOTE) msg("Running on AC power.", LOG_NOTE);
		config.max_speed = config.ac_max_speed;
		config.min_speed = config.ac_min_speed;
		config.up_threshold = config.ac_up_threshold;
		config.down_threshold = config.ac_down_threshold;
	}
}

#ifdef DEBUG
void print_config(void) {
	printf("This is the config I've read:\n");
	printf("Path to sysstat:\t%s\n", config.procstat);
	printf("Path to scaling governor:\t%s\n", config.sys->scaling_governor);
	printf("Path to scaling setspeed:\t%s\n", config.sys->scaling_setspeed);
	printf("Path to acpi throttling:\t%s\n", config.acpi->throttling);
	printf("Path to acpi ac state:\t%s\n", config.acpi->ac_state);
	printf("Run as daemon?\t%i\n", config.daemonize);
	printf("Loglevel:\t%i\n", config.loglevel);
	printf("Use cpufreq?\t%i\n", config.use_cpufreq);
	printf("Use acpi?\t%i\n", config.use_acpi);
	printf("Maximum CPU speed:\t%i\n", config.max_cpu_speed);
	printf("Minimum CPU speed:\t%i\n", config.min_cpu_speed);
	printf("Number of cpufreq states:\t%i\n", config.cpu_states);
	printf("Number of acpi states:\t%i\n", config.acpi_states);
	printf("Maximum Speed to use on AC:\t%i\n", config.ac_max_speed);
	printf("Minimum Speed to use on AC:\t%i\n", config.ac_min_speed);
	printf("Upper load threshold on AC:\t%i\n", config.ac_up_threshold);
	printf("Lower load threshold on AC:\t%i\n", config.ac_down_threshold);
	printf("Maximum Speed to use on battery:\t%i\n", config.ac_max_speed);
	printf("Minimum Speed to use on battery:\t%i\n", config.ac_min_speed);
	printf("Upper load threshold on battery:\t%i\n", config.ac_up_threshold);
	printf("Lower load threshold on battery:\t%i\n", config.ac_down_threshold);
}
#endif

void parse_cmdline(int argc, char *argv[]) {
	int i = 0;

	for(i = 1; i < argc; i++) {
		if(argv[i][0] != '-') continue;
		/* First, the short options. */
		switch(argv[i][1]) {
		case 'c': 
			conffile = argv[++i];
			break;
		case 'f':
			config.daemonize = 0;
			break;
		case 't':
			do_test = 1;
			break;
		case 'h':
			print_help();
			break;
		/* And now the GNU-style long options. */
		case '-':
			if(!strcmp(argv[i], "--config")) conffile = argv[++i];
			else if(!strcmp(argv[i], "--foreground")) config.daemonize = 0;
			else if(!strcmp(argv[i], "--test")) do_test = 1;
			else if(!strcmp(argv[i], "--help")) print_help();
			else print_help();
			break;
		default:
			print_help();
		}
	}
}

void print_help() {
	printf("Usage: autothrottle [-f] [-h] [-t] [-c /path/to/file]\n");
	printf("\t[--foreground] [--help] [--test] [--config /path/to/file]\n");
	printf("-f --foreground\t\t\tRun in foreground, don't daemonize.\n");
	printf("-h --help\t\t\tPrint this usage messageand exit.\n");
	printf("-t --test\t\t\tPerform a test of the specified configuration and exit.\n");
	printf("-c --config /path/to/file\tUse /path/to/file as configuration instead of /etc/autothrottle.conf.\n");

	exit(EXIT_SUCCESS);
}
