#ifndef _CONFIG_H_
#define _CONFIG_H_

#define OKVAL (0)
#define ERRVAL (-1)

typedef struct sysfs {
	char *scaling_governor;
	char *scaling_setspeed;
} sysfs;

typedef struct acpifs {
	char *throttling;
	char *ac_state;
} acpifs;

typedef struct options {
	char *procstat;
	sysfs *sys;
	acpifs *acpi;

	int daemonize;

	int loglevel;

	int use_cpufreq;
	int use_acpi;

	int max_cpu_speed;
	int min_cpu_speed;

	int cpu_states;
	int acpi_states;

	int max_speed;
	int min_speed;
	int up_threshold;
	int down_threshold;

	int ac_max_speed;
	int ac_min_speed;
	int ac_up_threshold;
	int ac_down_threshold;

	int bat_max_speed;
	int bat_min_speed;
	int bat_up_threshold;
	int bat_down_threshold;
} options;

/* our global config */

options config;
int do_test;

/* Wrapper for config parsing and checking function.
 * Arguments are the number of the commandline arguments and the array of them.
 * Returns OKVAL on success or ERRVAL on failure. */
int get_config(int argc, char *argv[]);

/* Parses the config file given as an argument for keyword = value pairs.
 * Returns OKVAL on success or ERRVAL on failure. */
int parse_config_file(char *name);

/* Checks if keyword is in the list of valid keywords and if so sets the value
 * of this option in the config struct to value. */
void set_value(char *keyword, char *value);

/* Verifies that the configuration options parsed are valid.
 * Returns OKVAL on success or ERRVAL on failure. */
int verify_config(void);

/* Checks whether we're running on battery or AC power and sets maximum speed,
 * minimum speed, high threshold and low threshold accordingly. */
void check_ac(void);

/* Prints the current configuration.
 * For debug use only. */
#ifdef DEBUG
void print_config(void);
#endif

/* Parses the commandline arguments.
 * Arguments are the number of commandline arguments and the array with them. */
void parse_cmdline(int argc, char *argv[]);

/* Prints a usage guide and terminates the program. */
void print_help();

#endif
