#include <stdio.h>
#include <stdlib.h>

#include "cpu.h"
#include "config.h"
#include "msg.h"

static unsigned long prev_idle = 0, prev_total = 0;

int enable_userspace_governor(void) {
	FILE *fd = NULL;
	fd = fopen(config.sys->scaling_governor, "w");
	if(fd == NULL) {
		msg("Unable to enable userspace scaling governor.", LOG_ERROR);
		return ERRVAL;
	}
	fprintf(fd, "userspace");
	fclose(fd);

	return OKVAL;
}

int set_speed(int newspeed) {
	FILE *fd = NULL;
	char *tmp = (char *) calloc(26, sizeof(char));
	fd = fopen(config.sys->scaling_setspeed, "w");
	if(fd == NULL) {
		msg("Unable to set new CPU speed via cpufreq interface.", LOG_ERROR);
		return ERRVAL;
	}
	fprintf(fd, "%i", newspeed);
	sprintf(tmp, "Changed speed to %i.", newspeed);
	if(config.loglevel == LOG_NOTE) msg(tmp, LOG_NOTE);
	fclose(fd);
	free(tmp);

	return OKVAL;
}

int get_speed(void) {
	FILE *fd = NULL;
	int cpu_speed = ERRVAL;
	
	fd = fopen(config.sys->scaling_setspeed, "r");
	if(fd == NULL) {
		msg("Unable to get current cpufreq throttling state.", LOG_ERROR);
		return ERRVAL;
	}
	fscanf(fd, "%i", &cpu_speed);
	fclose(fd);
	return cpu_speed;
}

double get_cpu_load(void) {
	FILE *fd = NULL;
	unsigned long delta_idle = 0, time_passed = 0, user = 0, nice = 0,\
		idle = 0, sys = 0, total = 0;

	/* default return value means error occurred */
	double load = -1.0;

	if((fd = fopen (config.procstat, "r")) == NULL) {
		msg("Unable to open system stats file.", LOG_ERROR);
		return -1.0;
	}

	/* read in the CPU stats that the Linux kernel has provided for us */
	if(!fscanf(fd, "cpu %lu %lu %lu %lu", &user, &nice, &sys, &idle)) {
		fclose(fd);
		msg("Unable to get CPU load from proc.", LOG_ERROR);
		return -1.0;
	}
	fclose(fd);

	/* compute time spent idling during this period */
	delta_idle = idle - prev_idle;

	/* compute time period */
	total = user + nice + sys + idle;
	time_passed = total - prev_total;

	/* Compute the % load from the time spent idling and the total time
	   in the sampling period */
	load = 100.0 - ((double) delta_idle / (double) time_passed * 100.0);

	/* retain the current counts for next time period */
	prev_idle = idle;
	prev_total = total;

	return load;
}
